#!/usr/bin/env python
"""
Find the closest points to a target location using the trace header database.

     Copyright (C) 2005  Kurt Schwehr

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

$Id: segysqldist.py,v 1.7 2005/08/22 20:21:20 schwehr Exp $

"""

# These values end up in the pydoc web page
__version__ = '$Revision: 1.7 $' # See man ident
__date__ = '$Date: 2005/08/22 20:21:20 $' # FIX: pull out just the date
__author__ = 'Kurt Schwehr'
__credits__ = """Lots of people for helping me with my thesis.
Especially Neal Driscoll and Lisa Tauxe!
"""

import mmap
import os
import struct	# Convert two and from binary formats
import sqlite
import sys

from optparse import OptionParser

# What version do we want to track?  This would be for cvs of this file
vcvs = "$Revision: 1.7 $"
vcvs = vcvs[vcvs.find(':')+2:]
vcvs = vcvs[:vcvs.find('$')-1]
version_cvs = vcvs
del (vcvs) # stop polution


from segy_version import VERSION

myparser = OptionParser(usage="%prog [options] segyfile1 segyfile2 ... segyfileN",
                      version="%prog "+VERSION+' - '+version_cvs)
myparser.add_option('-f','--file',dest='dbfilename',default='segy.db',
		    help='file for the sqlite database  [default: %default]')
myparser.add_option('-x','--x',dest='x', help='x coordinate (decimal degrees longitude)')
myparser.add_option('-y','--y',dest='y', help='y coordinate (decimal degrees latitude)')

myparser.add_option('-n','--num-shots',dest='N', default='1000', help='number of shots to find [default: %default].  More is slower')

myparser.add_option('-p','--point-file',dest='pointFile', default='point.dat', help='file to write out target x,y [default: %default].')

myparser.add_option('-d','--data-file',dest='dataFile', default=None, help='file to write out a subset of data searched [default: %default].')
myparser.add_option('-s','--sample-interval',dest='sampleInterval', default=1000, help='how many samples to skip when writing out --data-file [default: %default].')


myparser.add_option('-o','--output-file',dest='resultsFile', default='results-lonlat.dat', help='file to write out the closest points to [default: %default].')

myparser.add_option('-S','--do-not-skip',dest='ignoreSame', default=False, action='store_true',help='give this flag to keep samples in the list with the same xy values.  Some systems give the same GPS position over a number of shots.')


from closestN import ClosestPoints
    
if __name__ == "__main__":
    (options,args) = myparser.parse_args()


    if 0 != args:
        print args
        print "WARNING: there should be no extra args\n\n"
        # FIX: force no extra args... fail out
        
    if not options.x or not options.y:
        print "You must specify both x and y\n"
        myparser.print_help()
        sys.exit("Goodbye and good ridance")

    # FIX: range check on x and y

    dbName = options.dbfilename
    if not os.access(dbName,os.R_OK):
        print "ERROR: database does not exist...",dbName
        sys.exit("No shoes, no shirt, no service")
    cx = sqlite.connect (dbName)
    cu = cx.cursor()

    print "Starting SQL select.  This make take quite a while"
    cu.execute('SELECT X,Y,traceNumber,fileKey,Shotpoint FROM segyTrace')
    print "Finished SQL select"
    target = (options.x,options.y)

    o = open (options.pointFile,"w")
    o.write(options.x+" "+options.y+"\n")
    o.close()
    del(o)

    ########################################
    # N points
    ########################################
    i = 0
    if None != options.dataFile:
        o = open (options.dataFile,'w')
    print "# Beginning search"
    if 1<int(options.N):
        print "Using slower search for many points.  Suggest using -n 1 for faster"
        cPts = ClosestPoints(target,int (options.N))
        print "Inspecting data:"
        for row in cu.fetchall():
            i += 1
            if 0== i % 1000: print i
            x = float(row['X'])/3600.
            y = float(row['Y'])/3600.
            if None != options.dataFile:
                o.write(str(x)+' '+str(y)+'\n')  # FIX: maybe writeout a subset?
            point = (float(x),float(y))
            data = (row['fileKey'],row['Shotpoint'],row['traceNumber'])
            cPts.checkPoint(point,data,ignore=options.ignoreSame)
        if None != options.dataFile: del(o)

        o = open (options.resultsFile,'w')
        
        o.write('# segysqldist.py\n')
        o.write('# \n')
        o.write('# distance    - currently is in "degrees"... sort of\n')
        o.write('# traceNumber - the offset in the particular segy file\n')
        o.write('# fileKey     - use the segyFile table in *.db to translate\n')
        o.write('# \n')
        o.write('# longitude latitude "distance" fileKey Shotpoint traceNumber\n')
        for i in range(len(cPts.points)):
            p = cPts.points[i]
            d = cPts.pointData[i]
            s = ''
            for value in p:
                s += ("  %.6f" % value)
            for value in d:
                s += " " + str(value)
            s += '\n'
            o.write(s)
        del(o)

    ########################################
    # Single point - Much faster
    ########################################
    else:
        assert(1==int(options.N))
        bestFile = None
        bestShot = None
        bestX = None
        bestY = None
        bestDistance = 99999999999.

        # FIX: use a real distance function!!!!
        from closestN import greatCircleDistKm

        for row in cu.fetchall():
            i += 1
            if 0== i % 1000: print i
            x = float(row['X'])/3600.
            y = float(row['Y'])/3600.
            if None != options.dataFile:
                o.write(str(x)+' '+str(y)+'\n')  # FIX: maybe writeout a subset?
            dist = greatCircleDistKm((x,y),target)
            if dist < bestDistance:
                #print x,y,row['traceNumber'], row['fileKey'],row['Shotpoint'],target,dist
                bestDistance = dist
                bestShot = row['Shotpoint']
                bestX = x
                bestY = y
                bestFile = row['fileKey']

        print "BEST: ",bestFile,bestShot,bestX,bestY,bestDistance
        o = open (options.resultsFile,'w')
        o.write(str(bestX)+' '+str(bestY)+' '+str(bestDistance)+'\n')
        del(o)
    print "Thanks for playing"
