#!/usr/bin/env python
"""

Command line utility to help build maps with the Generic Mapping Tool
(GMT) package for file and trace header information stored in an
SQLite database.

     Copyright (C) 2005  Kurt Schwehr

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

$Id: segysqlgmt.py,v 1.6 2005/08/27 14:25:14 schwehr Exp $

http://www.python.org/doc/lib/module-optparse.html

http://schwehr.org/software/segy-py/
"""

# These values end up in the pydoc web page
__version__ = '$Revision: 1.6 $' # See man ident
__date__ = '$Date: 2005/08/27 14:25:14 $' # FIX: pull out just the date
__author__ = 'Kurt Schwehr'
__credits__ = """Lots of people for helping me, especially Neal
Driscoll and Lisa Tauxe."""

import sqlite
import os
import sys

from optparse import OptionParser

vcvs = "$Revision: 1.6 $"
vcvs = vcvs[vcvs.find(':')+2:]
vcvs = vcvs[:vcvs.find('$')-1]
version_cvs = vcvs
del (vcvs) # stop polution

# Track the version of segy-py.  If I use setup.py, which one gets imported?
from segy_version import VERSION

myparser = OptionParser(usage="%prog [options]",
                      version="%prog "+VERSION+' - '+version_cvs)
myparser.add_option('-d','--database',dest='dbfilename',default='segy.db',
                    help='Database containing imported segy file and trace header info [default: %default]')
myparser.add_option('-m','--min-file-key',dest='minFileKey',default=None, type='int',
                    help='Set this to start processing at a certain file index [default: %default]')
myparser.add_option('-M','--max-file-key',dest='maxFileKey',default=None, type='int',
                    help='Set this to end processing at a certain file index [default: %default]')

# FIX: Option to skip
myparser.add_option('-l','--line-file',dest='shotLineFilename',default='shot-locations.pstext',
                    help='Where to write the unique shot location line info [default: %default]')
myparser.add_option('-b','--batch-lines',dest='batchLineFiles',default=False,action='store_true',
                    help='Process all the files for lines together without picking up the pen [default: %default]')
# FIX: Option to force hemisphere to cope with xstar missing longitude missing negatives

# FIX: Option to skip
myparser.add_option('-s','--shot-label-file',dest='shotTextFilename',default='shot-labels.pstext',
                    help='Where to write the shot labels for every Nth shot [default: %default]')
myparser.add_option('-n','--nth-shot-label',dest='nthShotLabel',type='int',default='2000',
                    help='How many shots to skip before writing out a label [default: %default]')
myparser.add_option('-a','--label-angle',dest='shotLabelAngle',default='45',
                    help='How much to angle the text for each shot label [default: %default]')

myparser.add_option('-g','--gmt-script',dest='makeGmtScript',default=False,action='store_true',
                    help='Print out example GMT script code')

myparser.add_option('-S','--scaler-coord',dest='scalerCoord',default=False,action='store_true',
                    help='Divide by the scaler coord field in the trace header to get the true lat/lon')

myparser.add_option('-x','--flip-x',dest='flipX',default=False,action='store_true',
                    help='Multiply the x coordinate by -1')
myparser.add_option('-y','--flip-y',dest='flipY',default=False,action='store_true',
                    help='Multiply the x coordinate by -1')


myparser.add_option('-v','--verbose',action="count",dest='verbosity',default=0,
                    help='how information to give.  Specify multiple times to increase verbosity')

BOMBASTIC= 4
VERBOSE  = 3
TRACE    = 2
TERSE    = 1
ALWAYS   = 0


# FIX: Verbosity
# FIX: allow the code to detect location X,Y format.  Do not force /3600.0

def buildShotLines(db_cx, options, outFilename='monkey-boy.pstext',
                   minFileKey=None,maxFileKey=None,batchLineFiles=False,scalerCoord=False
                   , verbosity=TERSE):
    """FIX: describe

    FIX: when we see a shift in fileKey, emit a '>\n' to start a new line
    """
    print 'file key ranges = ',minFileKey,maxFileKey
    cu = db_cx.cursor()
    out = open(outFilename,'w')

    #
    # Refactor such that the sql query generation is pretty much shared!!!!
    # 

    if batchLineFiles:
        # FIX: Bad to hard code Seconds of Arc!"
        if scalerCoord:
            searchStr  = "SELECT (x/3600.0/ScalerCoord) as x , (y/3600.0/ScalerCoord) as y FROM segyTrace"
        else:
            searchStr  = "SELECT (x/3600.0) as x ,(y/3600.0) as y FROM segyTrace"
        if minFileKey or maxFileKey: searchStr += ' WHERE'
        if maxFileKey:
            searchStr += ' fileKey<='+str(maxFileKey)
        if minFileKey:
            if maxFileKey: searchStr += ' and'
            searchStr += ' fileKey>='+str(minFileKey)
        searchStr += ';'
        print 'FIX: verb  ',verbosity, TRACE
        if verbosity >= TRACE:
            print searchStr 
        cu.execute(searchStr)
        
        lastx = lasty = -1
        for row in cu.fetchall():
            x = row['x']
            y = row['y']
            if 0 == int(x) or 0 ==int(y):
                print "Skipping possible bad value"
                continue
            if 180<float(y) or -180>float(y) or -360 > float(x) or 360 < float(x):
                print "FIX: range error", x,y
            if options.flipX: x = -float(x)
            if options.flipY: y = -float(y)
            if lastx != x or lasty != y:
                # Only write out a line when we get a new location
                out.write(str(x)+" "+str(y)+"\n")
                lastx = x
                lasty = y

    else: # Do not batch.  Do each file as a separate SQL query.
        # FIX: need the SQL ORDER?
        # Faster to do the math in SQL or Python?
        for fileKey in range (int(minFileKey),int(maxFileKey)+1):
            print 'fileKey = ',fileKey
            if scalerCoord:
                searchStr  = "SELECT (x/3600.0/ScalerCoord) as x , (y/3600.0/ScalerCoord) as y FROM segyTrace"
            else:
                searchStr  = "SELECT (x/3600.0) as x ,(y/3600.0) as y FROM segyTrace"
            #searchStr  = "SELECT x/3600.0,y/3600.0 FROM segyTrace "
            searchStr += " WHERE fileKey="+str(fileKey)+";"
            # FIX: need order?  searchStr += "  ORDER BY shotpoint;"
            if verbosity >= TRACE:
                print searchStr 
            cu.execute(searchStr)
            lastx = lasty = -1
            for row in cu.fetchall():
                x = row['x']
                y = row['y']
                if 0 == int(x) or 0 ==int(y): continue
                if 180<float(y) or -180>float(y) or -360 > float(x) or 360 < float(x):
                    print "FIX: range error", x,y
                if options.flipX: x = -float(x)
                if options.flipY: y = -float(y)
                if lastx != x or lasty != y:
                    out.write(str(x)+" "+str(y)+"\n")
                    lastx = x
                    lasty = y
            out.write(">\n") # Start GMT psxy on a new line (-M option)
    out.close()
    return

def buildShotText(db_cx, outFilename='JohnYahYah.pstext',nthShotLabel=2000,shotLabelAngle=45,
                  minFileKey=None, maxFileKey=None):
    """
    FIX: describe

    FIX: add range checking on args
    """
    print "buildShotText",outFilename,nthShotLabel,shotLabelAngle
    cu = db_cx.cursor()
    out = open(outFilename,'w')
    searchStr  = "SELECT fileKey,shotpoint,x/3600.0,y/3600.0 FROM segyTrace "
    searchStr += "WHERE shotpoint % "+str(nthShotLabel)+"= 0" # FIX: need  ORDER BY shotpoint???. 
    if None != minFileKey: searchStr += " AND fileKey >= "+str(minFileKey)
    if None != maxFileKey: searchStr += " AND fileKey <= "+str(maxFileKey)
    searchStr += ";" # FIX: need  ORDER BY shotpoint???
    print searchStr
    cu.execute(searchStr)
    out = open(outFilename,'w')
    angle = str(shotLabelAngle)
    for row in cu.fetchall():
        x = str(row['x/3600.0'])
        y = str(row['y/3600.0'])
        shotpoint = str(row['shotpoint'])
        fileKey = str(row['fileKey'])
        out.write(x+" "+y+' 1 '+angle+' 1 LM '+shotpoint+'-'+fileKey+'\n')
    out.close()
    return

def printGmtSampleScript(options):
    """
    FIX: would be nice to give them more.
    """
    print "# Add the following lines to your gmt script or what was produced by mbm_grdplot"
    print
    print "# Draw the ship track for segy data"
    print "# -M is for multi seg files with '>' between"
    print "psxy "+options.shotLineFilename+" -W1/255/0/255 -M \\"
    print "        -J$MAP_PROJECTION$MAP_SCALE \\"
    print "        -R$MAP_REGION \\"
    print "        -K -O -V >> $PS_FILE"
    print
    print "# put a mark for the shot labels"
    print "psxy "+options.shotTextFilename+"  -Sc.05c   \\"
    print "        -G255/100/0 \\"
    print "        -J$MAP_PROJECTION$MAP_SCALE \\"
    print "        -R$MAP_REGION \\"
    print "        -K -O -V >> $PS_FILE"
    print
    print "# Label every nth shot"
    print "pstext "+options.shotTextFilename+" -G0/0/0 \\"
    print "        -J$MAP_PROJECTION$MAP_SCALE \\"
    print "        -R$MAP_REGION \\"
    print "        -K -O -V >> $PS_FILE"


if __name__ == "__main__":
    (options,args) = myparser.parse_args()

    if options.verbosity >= TRACE:
        print "Verbosity = ", options.verbosity

    # FIX: check for database first
    dbName = options.dbfilename
    cx = sqlite.connect(dbName)
    if not os.access(dbName,os.R_OK):
        print "ERROR: database does not exist...",dbName
        sys.exit("No shoes, no shirt, no service")

    specifiedFileKeys = False
    if None!=options.minFileKey or None!=options.maxFileKey: specifiedFileKeys = True

    if None==options.minFileKey or None==options.maxFileKey:
        cu = cx.cursor()
        cu.execute("SELECT MIN(fileKey),MAX(fileKey) FROM segyFile;")
        row = cu.fetchone()
        if None==options.minFileKey:
            options.minFileKey = int(row['MIN(fileKey)'])
        if None==options.maxFileKey:
            options.maxFileKey = int(row['MAX(fileKey)'])
        print "fileKey range found: ",options.minFileKey," to ", options.maxFileKey

    # FIX: make this call optional
    buildShotLines(cx, options, options.shotLineFilename,
                   minFileKey=options.minFileKey,maxFileKey=options.maxFileKey,
                   batchLineFiles=options.batchLineFiles
                   ,scalerCoord=options.scalerCoord
                   ,verbosity=options.verbosity)

    # FIX: make this call optional
    if (specifiedFileKeys):
        buildShotText(cx,options.shotTextFilename,options.nthShotLabel,options.shotLabelAngle,
                      minFileKey=options.minFileKey, maxFileKey=options.maxFileKey)
    else:
        buildShotText(cx,options.shotTextFilename,options.nthShotLabel,options.shotLabelAngle,
                      minFileKey=options.minFileKey, maxFileKey=options.maxFileKey)

    if options.makeGmtScript:
        printGmtSampleScript(options)

