"""
Angelo Secchi
secchi@sssup.it

http://193.205.80.238/~angelo/ibm.html

IBM System/370 mainframes and standard PC based on Intel
microprocessors use different internal storage for floating points
numbers. Here a brief description together with other sources of
informations.


IBM/370 Floating point format.  The single precision floating point is
represented by 32 bits. The first is the sign bit. From bit 1 to 7 you
find the exponent represented as a power of 16 in 'excess 64' notation
(giving a value range of 16**-65 <= M <= 16**63).From bit 8 to 31 is
the base-16 (hex) fraction left-shifted (normalized) 4 bits at a time
to where the first four bits (nibble) contain a numeric value on the
right of an imaginary radix-point (with the exponent raised
appropriately during the shift). For details see: IBM System/370 Model
168, Theory of Operation/Diagrams Manual, Volume 2, I-unit, Form
SY22-6932-0, IBM Corporation, Poughkeepsie, New York, April 1973.


IEEE-754 Floating point format.
The single precision floating point is represented by 32 bits. The first is the sign bit. From bit 1 to 7 you find the exponent represented as a power of 2 in 'excess 127' notation. From bit 8 to bit 31 you find the base-2 fraction left-shifted to where the first binary '1' is on the left of an imaginary binary-point (with the exponent raised appropriately during the shift) giving a precision of 24 bits stored in 23 bits. For details see: IEEE-754 Floating point format

Python Functions

Python is a simple language that makes available also to 'not programmer' (like I am) a powerful tool without investing huge amount of time (www.python.org). The material provided in this page comes from the Pyhton community and in particular from Howard Lightstone and Anton Vredegoor. Please notice that these two functions have been tested several times but only by myself.

Suppose you have a binary file, called data, created on an IBM/370 System with 223 equal fields where a single precision (4 bytes) float is stored and you want to convert it to IEEE-754 standard. Here the first sample code that uses the Python 'struct' module.
import struct
"""

def convert(ibm_float):
    i = struct.unpack('>I',ibm_float)[0]
    sign = [1,-1][bool(i & 0x100000000L)]
    characteristic = ((i >> 24) & 0x7f) - 64
    fraction = (i & 0xffffff)/float(0x1000000L)
    return sign*16**characteristic*fraction   


# inf=file('data','rb')
# dati=inf.read(892) # 223*4 

# for i in range(223):
#     index=i*4
#     print convert(dati[index: index+4])

"""
Here a different implementation of the same conversion procedure that consider the IBM/370 float as a set of 4 char binary strings (IBM/370-formatted).
"""

import struct

dividend=float(16**6)

def convert2(ibm_float):
    if ibm_float == 0:
        return 0.0
    istic,a,b,c=struct.unpack('>BBBB',ibm_float)
    if istic >= 128:
        sign= -1.0
        istic = istic - 128
    else:
        sign = 1.0
    mant= float(a<<16) + float(b<<8) +float(c)
    return sign* 16**(istic-64)*(mant/dividend)

#inf=file('data','rb')
#dati=inf.read(892) # 223*4 

#for i in range(223):
#    index=i*4
#    print convert2(dati[index: index+4])
